/* eslint-disable @typescript-eslint/no-explicit-any */

export type HTTPMETHOD =
  | 'GET'
  | 'POST'
  | 'PUT'
  | 'DELETE'
  | 'OPTIONS'
  | 'PATCH';
export type CustomParser<T> = (input: any) => T;
export type ZodParser<T> = {
  parseAsync: (input: any) => Promise<T>;
};

export type Parser<T = any> = ZodParser<T> | CustomParser<T>;
export type InferParser<T extends Parser> = T extends Parser<infer V>
  ? V
  : never;

export type Procedure<
  P extends Parser = Parser<any>,
  Input extends any[] = any[],
  C = Record<string, any>
> = {
  method: (
    context: Context<C>,
    ...args: Input
  ) => InferParser<P> | Promise<InferParser<P>>;
  parser?: P;
};

export type Context<T> = T & {
  fail: (status: number, ...data: any[]) => FailError;
};
export type Module = Record<string, (...args: any[]) => any | Module>;

export class FailError extends Error {
  status: number;
  data: any[];

  constructor(status: number, ...data: any[]) {
    super(`[Procedure Fail]: status ${status}`);

    this.status = status;
    this.data = data;
  }
}

export function createRepo<T extends Record<string, any>>(context: () => T) {
  const ctx = context();

  const fail = (status: number, ...data: any[]) => {
    return new FailError(status, ...data);
  };

  function module<M extends Module>(mod: M) {
    return Object.assign(mod, ctx);
  }

  function procedure<P extends Procedure<any, any[], T>>(
    proc: P
  ): P extends Procedure<infer R, infer S, T>
    ? (...args: S) => Promise<InferParser<R>>
    : Promise<any>;
  function procedure<P extends Procedure<any, any[], T>>(proc: P) {
    type ActionArguments = OmitFirstArg<typeof proc.method>;
    return async function (...args: ActionArguments) {
      const input = await proc.method({ ...ctx, fail }, ...args);
      if (typeof proc.parser === 'function') return proc.parser(input);
      if (proc.parser) return await proc.parser.parseAsync(input);
      return input;
    };
  }

  return { module, procedure };
}

/* -------------------------------------------------------------------------- */
type OmitFirstArg<F> = F extends (x: any, ...args: infer P) => any ? P : never;
