/* eslint-disable @typescript-eslint/no-explicit-any */
export interface Schema {
  [K: string]: Schema | string;
}

export type InferLiteralString<T, TD extends string> = T extends infer K
  ? K extends string
    ? K extends ''
      ? `${TD}`
      : `${K}`
    : `${TD}`
  : `${TD}`;

export type ScopePrefixer<T, S> = T extends infer K
  ? K extends string
    ? K extends ''
      ? ''
      : `${K}${InferLiteralString<S, '.'>}`
    : never
  : never;

export type UnionScopes<
  T,
  Prefix extends string = '',
  Spacer extends string = '.'
> = T extends Record<string, any>
  ? {
      [K in keyof T]: K extends string
        ? T[K] extends Record<string, any>
          ? UnionScopes<
              T[K],
              `${ScopePrefixer<Prefix, Spacer>}${K}`,
              InferLiteralString<Spacer, '.'>
            >
          : `${ScopePrefixer<Prefix, Spacer>}${K}`
        : never;
    }[keyof T]
  : Prefix;


export type PermitNode = {
  scope: string;
  granted: boolean;
  desc: string;
  toggle: () => void;
};

export type TransformSchema<T, TValue> = T extends Schema
  ? {
      [P in keyof T]: T[P] extends Schema
        ? TransformSchema<T[P], TValue>
        : TValue;
    }
  : TValue;

export type ScopeItem = {
  scope: string;
  toggle: () => void;
  desc?: string;
  parent: string;
  hasChild: boolean;
  value: {
    enabled: boolean;
    partialChild: boolean;
  };
};

export type AccessEditor = {
  toggle: (scope: string, asParent?: boolean) => void;
  getChild: (scope: string) => PermitNode[];
  enableAll: () => void;
  disableAll: () => void;
  enabled: readonly string[];
  disabled: readonly string[];
  scopes: readonly ScopeItem[];
};

export type SchemaOptions = {
  /** add prefix to scope. default to '' (none) */
  prefix: string;

  /** string to use for spacer for scope. default to '.' */
  spacer: string;
};
