/* eslint-disable @typescript-eslint/no-explicit-any */
import { Schema, TransformSchema } from './types';

export function traverseAccessControl<T extends Schema>(
  node: T,
  parent = '',
  spacer = '.',
  onLeaf: (node: string, scope: string, parent: string) => void = () =>
    undefined,
  onBranch: (node: Schema, scope: string, parent: string) => void = () =>
    undefined
) {
  for (const key in node) {
    if (Object.prototype.hasOwnProperty.call(node, key)) {
      const scope = parent !== '' ? [parent, key].join(spacer) : key;
      const current = node[key] as string | Schema;

      if (typeof current === 'string') {
        onLeaf(current, scope, parent);
      } else {
        onBranch(current, scope, parent);
        traverseAccessControl(current, scope, spacer, onLeaf, onBranch);
      }
    }
  }
}

export function transformAccessControl<A extends Schema, T>(
  nodes: A,
  parent: string,
  spacer = '.',
  onLeaf: (node: string, scope: string, parent: string) => T
): TransformSchema<A, T> {
  const traverse = <N extends string | Schema>(node: N, path: string) => {
    const scope = parent !== '' ? [parent, path].join(spacer) : path;
    if (typeof node === 'string') {
      const value = onLeaf(node, scope, path);
      return [path, value];
    } else {
      return [path, transformAccessControl(node, scope, spacer, onLeaf)];
    }
  };

  return Object.fromEntries(
    Object.entries(nodes).map(([key, value]) => traverse(value, key))
  );
}
